

// To compile: gcc -O3 -o d File-Name rijndael-alg-fst.h rijndael-alg-fst.c

#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <time.h>
#include "rijndael-alg-fst.h"

typedef unsigned char      UINT8;  /* 1 byte   */
typedef unsigned short     UINT16; /* 2 byte   */
typedef unsigned int       UINT32; /* 4 byte   */
typedef unsigned long long UINT64; /* 8 bytes  */
typedef unsigned long      UWORD;  /* Register */

#define DIGEST_KEY_LEN           16  /* DIGEST takes 16 bytes of external key */
#define MUL64(a,b) ((UINT64)((UINT64)(UINT32)(a) * (UINT64)(UINT32)(b)))
#define MUL32(d,k1,k2) ((UINT32) (((UINT32) MUL64(d,k1)) + ((UINT32) (MUL64(d,k2) >> 32))))
#define ADD32(a,b) ((UINT32) (((UINT32) a) + ((UINT32) ((UINT64) b >> 32))))

#define LOAD_UINT32_LITTLE(ptr)     (*(UINT32 *)(ptr))



#define DIGEST_OUTPUT_LEN       8 /* Allowable: 4, 8, 12, 16, 20, 32 bytes */
#define STREAMS (DIGEST_OUTPUT_LEN / 4) /* Number of times hash is applied  */
#define L1_KEY_LEN         1024       /* Internal key bytes                 */
#define L1_KEY_SHIFT         4       /* BYTES, Toeplitz key shift between streams */


/* DIGEST uses AES with 16 byte block and key lengths */
#define AES_BLOCK_LEN  16
#define AES_ROUNDS	   ((DIGEST_KEY_LEN / 4) + 6)
typedef UINT8          aes_int_key[AES_ROUNDS+1][4][4]; /* AES internal */
#define aes_encryption(in,out,int_key) \
	    rijndaelEncrypt((u32 *)(int_key), AES_ROUNDS, (u8 *)(in), (u8 *)(out))
#define aes_key_setup(key,int_key) \
	    rijndaelKeySetupEnc((u32 *)(int_key), (const unsigned char *)(key), \
	    DIGEST_KEY_LEN*8)

void kdf(void *buffer_ptr, aes_int_key key, UINT8 index, int nbytes)
{
    UINT8 in_buf[AES_BLOCK_LEN] = {0};
    UINT8 out_buf[AES_BLOCK_LEN];
    UINT8 *dst_buf = (UINT8 *)buffer_ptr;
    int i;
    
    /* Setup the initial value */
    in_buf[AES_BLOCK_LEN-9] = index;
    in_buf[AES_BLOCK_LEN-1] = i = 1;
        
    while (nbytes >= AES_BLOCK_LEN) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,AES_BLOCK_LEN);
        in_buf[AES_BLOCK_LEN-1] = ++i;
        nbytes -= AES_BLOCK_LEN;
        dst_buf += AES_BLOCK_LEN;
    }
    if (nbytes) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,nbytes);
    }
}

#if (DIGEST_OUTPUT_LEN == 4) // ONE 32-BIT WORDS OR 32 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8;
    
    h = *((UINT32 *)hp);
    k0 = *(k+0);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

        k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); k4 = *(k+4); 
	k5 = *(k+5); k6 = *(k+6); k7 = *(k+7); k8 = *(k+8);

	h+= MUL32(d0,k0,k1);
	h+= MUL32(d1,k1,k2);
	h+= MUL32(d2,k2,k3);
	h+= MUL32(d3,k3,k4);
	h+= MUL32(d4,k4,k5);
	h+= MUL32(d5,k5,k6);
	h+= MUL32(d6,k6,k7);
	h+= MUL32(d7,k7,k8);

	k0 = k8;

        d += 8;
        k += 8;
    } while (--c);
  *((UINT32 *)hp) = h;
}

#elif (DIGEST_OUTPUT_LEN == 8) // TWO 32-BIT WORDS OR 32 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
  UINT32 h1, h2;
  UWORD c = dlen / 32;
  UINT32 *k = (UINT32 *)kp;
  UINT32 *d = (UINT32 *)dp;
  UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
  UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8,k9;

  h1 = *((UINT32 *)hp);
  h2 = *((UINT32 *)hp + 1);

  k0 = *(k+0); k1 = *(k+1);
  do {
    d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
    d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
    d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
    d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

    k2 = *(k+2); k3 = *(k+3); k4 = *(k+4); k5 = *(k+5); 
    k6 = *(k+6); k7 = *(k+7); k8 = *(k+8); k9 = *(k+9);

    h1+= MUL32(d0,k0,k1); h2+= MUL32(d0,k1,k2); 
    h1+= MUL32(d1,k1,k2); h2+= MUL32(d1,k2,k3);
    h1+= MUL32(d2,k2,k3); h2+= MUL32(d2,k3,k4);
    h1+= MUL32(d3,k3,k4); h2+= MUL32(d3,k4,k5);
    h1+= MUL32(d4,k4,k5); h2+= MUL32(d4,k5,k6);
    h1+= MUL32(d5,k5,k6); h2+= MUL32(d5,k6,k7);
    h1+= MUL32(d6,k6,k7); h2+= MUL32(d6,k7,k8);
    h1+= MUL32(d7,k7,k8); h2+= MUL32(d7,k8,k9);

    k0 = k8; k1 = k9;

    d += 8;
    k += 8;
  } while (--c);
  ((UINT32 *)hp)[0] = h1;
  ((UINT32 *)hp)[1] = h2;
}

#elif (DIGEST_OUTPUT_LEN == 12) // THREE 32-BIT WORDS OR 96 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h1,h2,h3;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6;
    
    h1 = *((UINT32 *)hp);
    h2 = *((UINT32 *)hp + 1);
    h3 = *((UINT32 *)hp + 2);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k3 = *(k+3); k4 = *(k+4); k5 = *(k+5); k6 = *(k+6);

      h1+= MUL32(d0,k0,k1); h2+= MUL32(d0,k1,k2); h3+= MUL32(d0,k2,k3);
      h1+= MUL32(d1,k1,k2); h2+= MUL32(d1,k2,k3); h3+= MUL32(d1,k3,k4);
      h1+= MUL32(d2,k2,k3); h2+= MUL32(d2,k3,k4); h3+= MUL32(d2,k4,k5);
      h1+= MUL32(d3,k3,k4); h2+= MUL32(d3,k4,k5); h3+= MUL32(d3,k5,k6);

      k0 = k4; k1 = k5; k2 = k6;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT32 *)hp)[0] = h1;
    ((UINT32 *)hp)[1] = h2;
    ((UINT32 *)hp)[2] = h3;
}

#elif (DIGEST_OUTPUT_LEN == 16) // FOUR 32-BIT WORDS OR 128 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h1,h2,h3,h4;
    UINT64 m1,m2,m3;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7;
    
    h1 = *((UINT32 *)hp);
    h2 = *((UINT32 *)hp + 1);
    h3 = *((UINT32 *)hp + 2);
    h4 = *((UINT32 *)hp + 3);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k4 = *(k+4);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);

      /*
      h1+= MUL32(d0,k0,k1); h2+= MUL32(d0,k1,k2); h3+= MUL32(d0,k2,k3); h4+= MUL32(d0,k3,k4);
      h1+= MUL32(d1,k1,k2); h2+= MUL32(d1,k2,k3); h3+= MUL32(d1,k3,k4); h4+= MUL32(d1,k4,k5);
      h1+= MUL32(d2,k2,k3); h2+= MUL32(d2,k3,k4); h3+= MUL32(d2,k4,k5); h4+= MUL32(d2,k5,k6);
      h1+= MUL32(d3,k3,k4); h2+= MUL32(d3,k4,k5); h3+= MUL32(d3,k5,k6); h4+= MUL32(d3,k6,k7);
      */

      m1 = MUL64(d0,k1); m2 = MUL64(d0,k2); m3 = MUL64(d0,k3);
      h1+= ADD32(MUL64(d0,k0),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3);
      h4+= ADD32(m3,MUL64(d0,k4));

      m1 = MUL64(d1,k2); m2 = MUL64(d1,k3); m3 = MUL64(d1,k4);
      h1+= ADD32(MUL64(d1,k1),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3);
      h4+= ADD32(m3,MUL64(d1,k5));

      m1 = MUL64(d2,k3); m2 = MUL64(d2,k4); m3 = MUL64(d2,k5);
      h1+= ADD32(MUL64(d2,k2),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3);
      h4+= ADD32(m3,MUL64(d2,k6));

      m1 = MUL64(d3,k4); m2 = MUL64(d3,k5); m3 = MUL64(d3,k6);
      h1+= ADD32(MUL64(d3,k3),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3);
      h4+= ADD32(m3,MUL64(d3,k7));

      k0 = k4; k1 = k5; k2 = k6; k3 = k7;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT32 *)hp)[0] = h1;
    ((UINT32 *)hp)[1] = h2;
    ((UINT32 *)hp)[2] = h3;
    ((UINT32 *)hp)[3] = h4;
}

#elif (DIGEST_OUTPUT_LEN == 20) // Five 32-BIT WORDS OR 160 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h1,h2,h3,h4,h5;
    UINT64 m1,m2,m3,m4;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8;
    
    h1 = *((UINT32 *)hp);
    h2 = *((UINT32 *)hp + 1);
    h3 = *((UINT32 *)hp + 2);
    h4 = *((UINT32 *)hp + 3);
    h5 = *((UINT32 *)hp + 4);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); k4 = *(k+4);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k5 = *(k+5); k6 = *(k+6); k7 = *(k+7); k8 = *(k+8);

      m1 = MUL64(d0,k1); m2 = MUL64(d0,k2); m3 = MUL64(d0,k3); m4 = MUL64(d0,k4); 
      h1+= ADD32(MUL64(d0,k0),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,MUL64(d0,k5));

      m1 = MUL64(d1,k2); m2 = MUL64(d1,k3); m3 = MUL64(d1,k4); m4 = MUL64(d1,k5);
      h1+= ADD32(MUL64(d1,k1),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,MUL64(d1,k6));

      m1 = MUL64(d2,k3); m2 = MUL64(d2,k4); m3 = MUL64(d2,k5); m4 = MUL64(d2,k6);
      h1+= ADD32(MUL64(d2,k2),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,MUL64(d2,k7));

      m1 = MUL64(d3,k4); m2 = MUL64(d3,k5); m3 = MUL64(d3,k6); m4 = MUL64(d3,k7);
      h1+= ADD32(MUL64(d3,k3),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,MUL64(d3,k8));

      k0 = k4; k1 = k5; k2 = k6; k3 = k7; k4 = k8;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT32 *)hp)[0] = h1;
    ((UINT32 *)hp)[1] = h2;
    ((UINT32 *)hp)[2] = h3;
    ((UINT32 *)hp)[3] = h4;
    ((UINT32 *)hp)[4] = h5;
}

#elif (DIGEST_OUTPUT_LEN == 32) // EIGHT 32-BIT WORDS OR 256 BITS

static void digest_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h1,h2,h3,h4,h5,h6,h7,h8;
    UINT64 m1,m2,m3,m4,m5,m6,m7;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8,k9,k10,k11;
    
    h1 = *((UINT32 *)hp);
    h2 = *((UINT32 *)hp + 1);
    h3 = *((UINT32 *)hp + 2);
    h4 = *((UINT32 *)hp + 3);
    h5 = *((UINT32 *)hp + 4);
    h6 = *((UINT32 *)hp + 5);
    h7 = *((UINT32 *)hp + 6);
    h8 = *((UINT32 *)hp + 7);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); 
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); k11 = *(k+11);

      m1 = MUL64(d0,k1); m2 = MUL64(d0,k2); m3 = MUL64(d0,k3); m4 = MUL64(d0,k4);
      m5 = MUL64(d0,k5); m6 = MUL64(d0,k6); m7 = MUL64(d0,k7);
      h1+= ADD32(MUL64(d0,k0),m1);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,m5); h6+= ADD32(m5,m6); h7+= ADD32(m6,m7);
      h8+= ADD32(m7,MUL64(d0,k8));
      
      m1 = MUL64(d1,k2); m2 = MUL64(d1,k3); m3 = MUL64(d1,k4); m4 = MUL64(d1,k5); 
      m5 = MUL64(d1,k6); m6 = MUL64(d1,k7); m7 = MUL64(d1,k8);
      h1+= MUL64(d1,k1) +(m1 >> 32);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,m5); h6+= ADD32(m5,m6); h7+= ADD32(m6,m7);
      h8+= ADD32(m7,MUL64(d1,k9));

      m1 = MUL64(d2,k3); m2 = MUL64(d2,k4); m3 = MUL64(d2,k5); m4 = MUL64(d2,k6);
      m5 = MUL64(d2,k7); m6 = MUL64(d2,k8); m7 = MUL64(d2,k9);
      h1+= MUL64(d2,k2) +(m1 >> 32);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,m5); h6+= ADD32(m5,m6); h7+= ADD32(m6,m7);
      h8+= ADD32(m7,MUL64(d2,k10));

      m1 = MUL64(d3,k4); m2 = MUL64(d3,k5); m3 = MUL64(d3,k6); m4 = MUL64(d3,k7);
      m5 = MUL64(d3,k8); m6 = MUL64(d3,k9); m7 = MUL64(d3,k10);
      h1+= MUL64(d3,k3) +(m1 >> 32);
      h2+= ADD32(m1,m2); h3+= ADD32(m2,m3); h4+= ADD32(m3,m4);
      h5+= ADD32(m4,m5); h6+= ADD32(m5,m6); h7+= ADD32(m6,m7);
      h8+= ADD32(m7,MUL64(d3,k11));

      k0 = k4; k1 = k5; k2 = k6;  k3 = k7; 
      k4 = k8; k5 = k9; k6 = k10; k7 = k11;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT32 *)hp)[0] = h1;
    ((UINT32 *)hp)[1] = h2;
    ((UINT32 *)hp)[2] = h3;
    ((UINT32 *)hp)[3] = h4;
    
    ((UINT32 *)hp)[4] = h5;
    ((UINT32 *)hp)[5] = h6;
    ((UINT32 *)hp)[6] = h7;
    ((UINT32 *)hp)[7] = h8;
    
}


#endif

static double run_cpb_test(UINT8 digest_key[], int nbytes, char *data_ptr,
                           int data_len, double hz)
{
    clock_t ticks;
    double secs;
    unsigned long iters_per_tag;
    unsigned long tag_iters, i, no_iters;
    unsigned long total_mbs;

    int len;
    char *input;
    UINT8 digest_result[STREAMS*sizeof(UINT32)];

    total_mbs = 2500;
    tag_iters = (total_mbs * 1024 * 1024) / (nbytes) + 1;    
    iters_per_tag = nbytes / data_len;
    no_iters = iters_per_tag * tag_iters;

    //===================================================                
    
    len = data_len;
    input = data_ptr;
    while (len >= L1_KEY_LEN) {
      digest_aux(digest_key, input, digest_result, L1_KEY_LEN);
      len -= L1_KEY_LEN;
      input += L1_KEY_LEN;
    } 
                
    //==================================================

    ticks = clock();
    for(i = 0; i < no_iters; i++){
	len = data_len;
	input = data_ptr;

	while (len >= L1_KEY_LEN) {
	  digest_aux(digest_key, input, digest_result, L1_KEY_LEN);
	  len -= L1_KEY_LEN;
	  input += L1_KEY_LEN;
	}
    }
    ticks = clock() - ticks;
    secs = (double)ticks / CLOCKS_PER_SEC;

    return (secs * (hz/(tag_iters * nbytes)));
}

int main(void)
{
    aes_int_key prf_key;
    UINT8  digest_key [L1_KEY_LEN + L1_KEY_SHIFT * STREAMS];
    char *data_ptr;
    int data_len, i;
    double hz;
    double cpb;

    //int length_pts = 1024*1024; 
    int length_pts[] = {8*1024,32*1024,256*1024,1024*1024};

    hz = ((double)1e9);
    data_len = 8192;

    data_ptr = (char *)malloc(data_len + 16);

    for (i = 0; i < data_len; i++)
        data_ptr[i] = (i*i) % 128;

    aes_key_setup("abcdefghijklmnopqrstuvwxyz",prf_key);
    kdf(digest_key, prf_key, 1, sizeof(digest_key));
        
    //cpb = run_cpb_test(digest_key, length_pts, data_ptr, data_len, hz);
    //printf("\nAuthenticating %8d byte messages: %5.2f cpb.\n", length_pts, cpb);

    printf("\nSpeed of %d-bit output digest(k,m) \n",DIGEST_OUTPUT_LEN * 8);
    if (sizeof(length_pts) > 0) {
        for (i = 0; i < (int)(sizeof(length_pts)/sizeof(int)); i++) {
            cpb = run_cpb_test(digest_key, length_pts[i], data_ptr, data_len, hz);
            printf("Digesting %8d byte messages: %5.2f cpb.\n",
                                   length_pts[i], cpb);
        }
    }
    printf("\n");
}




