
// To compile: gcc -O3 -o all File-Name rijndael-alg-fst.h rijndael-alg-fst.c
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <time.h>
#include "rijndael-alg-fst.h"

typedef unsigned char      UINT8;  /* 1 byte   */
typedef unsigned short     UINT16; /* 2 byte   */
typedef unsigned int       UINT32; /* 4 byte   */
typedef unsigned long long UINT64; /* 8 bytes  */
typedef unsigned long      UWORD;  /* Register */

#define KEY_LEN             16  /* 16 bytes of external key */
#define L1_KEY_LEN         1024       /* Internal key bytes */
#define L1_KEY_SHIFT         4    
#define AES_BLOCK_LEN  16
#define AES_ROUNDS	   ((KEY_LEN / 4) + 6)
typedef UINT8          aes_int_key[AES_ROUNDS+1][4][4]; /* AES internal */
#define aes_encryption(in,out,int_key) \
	    rijndaelEncrypt((u32 *)(int_key), AES_ROUNDS, (u8 *)(in), (u8 *)(out))
#define aes_key_setup(key,int_key) \
	    rijndaelKeySetupEnc((u32 *)(int_key), (const unsigned char *)(key), \
	    KEY_LEN*8)

void kdf(void *buffer_ptr, aes_int_key key, UINT8 index, int nbytes)
{
    UINT8 in_buf[AES_BLOCK_LEN] = {0};
    UINT8 out_buf[AES_BLOCK_LEN];
    UINT8 *dst_buf = (UINT8 *)buffer_ptr;
    int i;
    
    in_buf[AES_BLOCK_LEN-9] = index;
    in_buf[AES_BLOCK_LEN-1] = i = 1;
        
    while (nbytes >= AES_BLOCK_LEN) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,AES_BLOCK_LEN);
        in_buf[AES_BLOCK_LEN-1] = ++i;
        nbytes -= AES_BLOCK_LEN;
        dst_buf += AES_BLOCK_LEN;
    }
    if (nbytes) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,nbytes);
    }
}

#define MUL32(d,k1,k2) ((UINT32) (((UINT32) MUL64(d,k1)) + ((UINT32) (MUL64(d,k2) >> 32))))
#define MUL64(a,b) ((UINT64)((UINT64)(UINT32)(a) * (UINT64)(UINT32)(b)))
#define LOAD_UINT32_LITTLE(ptr)     (*(UINT32 *)(ptr))

static void digest(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT32 h;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3, k0,k1,k2,k3,k4;
    
    h = *((UINT32 *)hp);
    k0 = *(k+0);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); k4 = *(k+4); 
	h+= MUL32(d0,k0,k1); h+= MUL32(d1,k1,k2);
	h+= MUL32(d2,k2,k3); h+= MUL32(d3,k3,k4);
	k0 = k4; d += 4; k += 4;
    } while (--c);
  *((UINT32 *)hp) = h;
}

void mmh_mod_P(void *hp)
{
  UINT64 h, a, b, z, p;
  long y, c, d;

  p = (0x00010000UL) + 15;
  h = *((UINT64 *)hp);
  a = h >> 32; b = h & 0x0000ffffUL;
  y = b - 15 * a;

  if(y < 0){
    c = y >> 32; d = y & 0x0000ffffUL;
    z = d - 15 * c;
    if(z >= p){
      z = z - p;
    }
    *((UINT32 *)hp) = z;
  }else{
    *((UINT32 *)hp) = y;
  }
}

static void mmh(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3, k0,k1,k2,k3;

    h = *((UINT64 *)hp);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
        h+= MUL64(d0,k0); h+= MUL64(d1,k1); h+= MUL64(d2,k2); h+= MUL64(d3,k3);
        d += 4; k += 4;
    } while (--c);
  *((UINT64 *)hp) = h;
}

static void nh(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7, k0,k1,k2,k3,k4,k5,k6,k7;
    
    h = *((UINT64 *)hp);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);
        k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
        k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);
        h += MUL64((k0 + d0), (k1 + d1));
        h += MUL64((k2 + d2), (k3 + d3));
        h += MUL64((k4 + d4), (k5 + d5));
        h += MUL64((k6 + d6), (k7 + d7));
        d += 8; k += 8;
    } while (--c);
  *((UINT64 *)hp) = h;
}

static void run_cpb_test(UINT8 key[], int nbytes, char *data_ptr, int data_len, double hz)
{
    clock_t ticks;
    double secs, cpb;
    unsigned long iters_per_tag;
    unsigned long tag_iters, i, no_iters;
    unsigned long total_mbs;
    int len;
    char *input;
    UINT8 result32[sizeof(UINT32)];
    UINT8 result64[sizeof(UINT64)];

    total_mbs = 2500;
    tag_iters = (total_mbs * 1024 * 1024) / (nbytes) + 1;    
    iters_per_tag = nbytes / data_len;
    no_iters = iters_per_tag * tag_iters;

    len = data_len;
    input = data_ptr;
    while (len >= L1_KEY_LEN) {
      digest(key, input, result32, L1_KEY_LEN);
      mmh(key, input, result64, L1_KEY_LEN);
      nh(key, input, result64, L1_KEY_LEN);
      len -= L1_KEY_LEN;
      input += L1_KEY_LEN;
    }

    //run speed test of digest
    ticks = clock();
    for(i = 0; i < no_iters; i++){
      len = data_len;	input = data_ptr;
      while (len >= L1_KEY_LEN) {
	digest(key, input, result32, L1_KEY_LEN);
	len -= L1_KEY_LEN; input += L1_KEY_LEN;
      }
    }
    ticks = clock() - ticks;
    secs = (double) ticks / CLOCKS_PER_SEC;
    cpb = secs * (hz/(tag_iters * nbytes));
    printf("\n Digest: %5.2f cpb.", cpb);

    //run speed test of MMH
    ticks = clock();
    for(i = 0; i < no_iters; i++){
      len = data_len;	input = data_ptr;
      while (len >= L1_KEY_LEN) {
	mmh(key, input, result64, L1_KEY_LEN);
	len -= L1_KEY_LEN; input += L1_KEY_LEN;
      }
      mmh_mod_P(result64);
    }
    ticks = clock() - ticks;
    secs = (double) ticks / CLOCKS_PER_SEC;
    cpb = secs * (hz/(tag_iters * nbytes));
    printf("\n MMH:\t %5.2f cpb.", cpb);

    //run speed test of NH
    ticks = clock();
    for(i = 0; i < no_iters; i++){
      len = data_len;	input = data_ptr;
      while (len >= L1_KEY_LEN) {
	nh(key, input, result64, L1_KEY_LEN);
	len -= L1_KEY_LEN; input += L1_KEY_LEN;
      }
    }
    ticks = clock() - ticks;
    secs = (double) ticks / CLOCKS_PER_SEC;
    cpb = secs * (hz/(tag_iters * nbytes));
    printf("\n NH:\t %5.2f cpb.\n", cpb);
}

int main(void)
{
    aes_int_key prf_key;
    UINT8 key [L1_KEY_LEN + L1_KEY_SHIFT];
    char *data_ptr;
    int data_len, i;
    double hz;
    int length_pts = 8*1024; 
    
    hz = ((double)1e9);
    data_len = 8192;

    data_ptr = (char *)malloc(data_len + 16);
    for (i = 0; i < data_len; i++)
        data_ptr[i] = (i*i) % 128;

    aes_key_setup("abcdefghijklmnopqrstuvwxyz",prf_key);
    kdf(key, prf_key, 1, sizeof(key));
    printf("\n Hashing %8d byte messages.", data_len);
    run_cpb_test(key, length_pts, data_ptr, data_len, hz);
    printf("\n");
}




