

// To compile: gcc -O3 -o mmh File-Name rijndael-alg-fst.h rijndael-alg-fst.c

#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <time.h>
#include "rijndael-alg-fst.h"

typedef unsigned char      UINT8;  /* 1 byte   */
typedef unsigned short     UINT16; /* 2 byte   */
typedef unsigned int       UINT32; /* 4 byte   */
typedef unsigned long long UINT64; /* 8 bytes  */
typedef unsigned long      UWORD;  /* Register */

#define MMH_KEY_LEN           16  /* MMH takes 16 bytes of external key */
#define MUL64(a,b) ((UINT64)((UINT64)(UINT32)(a) * (UINT64)(UINT32)(b)))
#define LOAD_UINT32_LITTLE(ptr)     (*(UINT32 *)(ptr))



#define MMH_OUTPUT_LEN       4    /* Allowable: 4, 8, 12, 16, 20, 32 bytes */
#define STREAMS (MMH_OUTPUT_LEN / 4) /* Number of times hash is applied  */
#define L1_KEY_LEN         1024       /* Internal key bytes                 */
#define L1_KEY_SHIFT         4       /* BYTES, Toeplitz key shift between streams */


/* MMH uses AES with 16 byte block and key lengths */
#define AES_BLOCK_LEN  16
#define AES_ROUNDS	   ((MMH_KEY_LEN / 4) + 6)
typedef UINT8          aes_int_key[AES_ROUNDS+1][4][4]; /* AES internal */
#define aes_encryption(in,out,int_key) \
	    rijndaelEncrypt((u32 *)(int_key), AES_ROUNDS, (u8 *)(in), (u8 *)(out))
#define aes_key_setup(key,int_key) \
	    rijndaelKeySetupEnc((u32 *)(int_key), (const unsigned char *)(key), \
	    MMH_KEY_LEN*8)

void kdf(void *buffer_ptr, aes_int_key key, UINT8 index, int nbytes)
{
    UINT8 in_buf[AES_BLOCK_LEN] = {0};
    UINT8 out_buf[AES_BLOCK_LEN];
    UINT8 *dst_buf = (UINT8 *)buffer_ptr;
    int i;
    
    /* Setup the initial value */
    in_buf[AES_BLOCK_LEN-9] = index;
    in_buf[AES_BLOCK_LEN-1] = i = 1;
        
    while (nbytes >= AES_BLOCK_LEN) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,AES_BLOCK_LEN);
        in_buf[AES_BLOCK_LEN-1] = ++i;
        nbytes -= AES_BLOCK_LEN;
        dst_buf += AES_BLOCK_LEN;
    }
    if (nbytes) {
        aes_encryption(in_buf, out_buf, key);
        memcpy(dst_buf,out_buf,nbytes);
    }
}

#if (MMH_OUTPUT_LEN == 4) // ONE 32-BIT WORDS OR 32 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h;
    UWORD c = dlen / 32;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7;
    
    h = *((UINT64 *)hp);
    do {
        d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
        d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
        d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
        d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

        k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);
        k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);

        h+= MUL64(d0,k0);
        h+= MUL64(d1,k1);  
        h+= MUL64(d2,k2); 
        h+= MUL64(d3,k3);  
        h+= MUL64(d4,k4); 
        h+= MUL64(d5,k5); 
        h+= MUL64(d6,k6); 
        h+= MUL64(d7,k7);  

        d += 8;
        k += 8;
    } while (--c);
  *((UINT64 *)hp) = h;
}

#elif (MMH_OUTPUT_LEN == 8) // TWO 32-BIT WORDS OR 32 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
  UINT64 h1, h2;
  UWORD c = dlen / 32;
  UINT32 *k = (UINT32 *)kp;
  UINT32 *d = (UINT32 *)dp;
  UINT32 d0,d1,d2,d3,d4,d5,d6,d7;
  UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8;

  h1 = *((UINT64 *)hp);
  h2 = *((UINT64 *)hp + 1);

  k0 = *(k+0);
  do {
    d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
    d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);
    d4 = LOAD_UINT32_LITTLE(d+4); d5 = LOAD_UINT32_LITTLE(d+5);
    d6 = LOAD_UINT32_LITTLE(d+6); d7 = LOAD_UINT32_LITTLE(d+7);

    k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); k4 = *(k+4); 
    k5 = *(k+5); k6 = *(k+6); k7 = *(k+7); k8 = *(k+8); 

    h1+= MUL64(d0,k0);     h2+= MUL64(d0,k1);  
    h1+= MUL64(d1,k1);     h2+= MUL64(d1,k2); 
    h1+= MUL64(d2,k2);     h2+= MUL64(d2,k3);  
    h1+= MUL64(d3,k3);     h2+= MUL64(d3,k4); 
    h1+= MUL64(d4,k4);     h2+= MUL64(d4,k5);  
    h1+= MUL64(d5,k5);     h2+= MUL64(d5,k6); 
    h1+= MUL64(d6,k6);     h2+= MUL64(d6,k7);  
    h1+= MUL64(d7,k7);     h2+= MUL64(d7,k8);  

    k0 = k8;

    d += 8;
    k += 8;
  } while (--c);
  ((UINT64 *)hp)[0] = h1;
  ((UINT64 *)hp)[1] = h2;
}

#elif (MMH_OUTPUT_LEN == 12) // THREE 32-BIT WORDS OR 96 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);

    k0 = *(k+0); k1 = *(k+1);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k2 = *(k+2); k3 = *(k+3); k4 = *(k+4); k5 = *(k+5); 

      h1+= MUL64(d0,k0); h2+= MUL64(d0,k1); h3+= MUL64(d0,k2); 
      h1+= MUL64(d1,k1); h2+= MUL64(d1,k2); h3+= MUL64(d1,k3); 
      h1+= MUL64(d2,k2); h2+= MUL64(d2,k3); h3+= MUL64(d2,k4); 
      h1+= MUL64(d3,k3); h2+= MUL64(d3,k4); h3+= MUL64(d3,k5); 

      k0 = k4; k1 = k5;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
}

#elif (MMH_OUTPUT_LEN == 16) // FOUR 32-BIT WORDS OR 128 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3,h4;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k3 = *(k+3); k4 = *(k+4); k5 = *(k+5); k6 = *(k+6);

      h1+= MUL64(d0,k0); h2+= MUL64(d0,k1); h3+= MUL64(d0,k2); h4+= MUL64(d0,k3); 
      h1+= MUL64(d1,k1); h2+= MUL64(d1,k2); h3+= MUL64(d1,k3); h4+= MUL64(d1,k4); 
      h1+= MUL64(d2,k2); h2+= MUL64(d2,k3); h3+= MUL64(d2,k4); h4+= MUL64(d2,k5);
      h1+= MUL64(d3,k3); h2+= MUL64(d3,k4); h3+= MUL64(d3,k5); h4+= MUL64(d3,k6);

      k0 = k4; k1 = k5; k2 = k6;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
}

#elif (MMH_OUTPUT_LEN == 20) // Five 32-BIT WORDS OR 160 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3,h4,h5;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);
    h5 = *((UINT64 *)hp + 4);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k4 = *(k+4); k5 = *(k+5); k6 = *(k+6); k7 = *(k+7);

      h1+= MUL64(d0,k0); h2+= MUL64(d0,k1); h3+= MUL64(d0,k2); h4+= MUL64(d0,k3); h5+= MUL64(d0,k4); 
      h1+= MUL64(d1,k1); h2+= MUL64(d1,k2); h3+= MUL64(d1,k3); h4+= MUL64(d1,k4); h5+= MUL64(d1,k5);
      h1+= MUL64(d2,k2); h2+= MUL64(d2,k3); h3+= MUL64(d2,k4); h4+= MUL64(d2,k5); h5+= MUL64(d2,k6);
      h1+= MUL64(d3,k3); h2+= MUL64(d3,k4); h3+= MUL64(d3,k5); h4+= MUL64(d3,k6); h5+= MUL64(d3,k7);

      k0 = k4; k1 = k5; k2 = k6; k3 = k7;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
    ((UINT64 *)hp)[4] = h5;
}

#elif (MMH_OUTPUT_LEN == 32) // EIGHT 32-BIT WORDS OR 256 BITS

static void mmh_aux(void *kp, void *dp, void *hp, UINT32 dlen)
{
    UINT64 h1,h2,h3,h4,h5,h6,h7,h8;
    UWORD c = dlen / 16;
    UINT32 *k = (UINT32 *)kp;
    UINT32 *d = (UINT32 *)dp;
    UINT32 d0,d1,d2,d3;
    UINT32 k0,k1,k2,k3,k4,k5,k6,k7,k8,k9,k10;
    
    h1 = *((UINT64 *)hp);
    h2 = *((UINT64 *)hp + 1);
    h3 = *((UINT64 *)hp + 2);
    h4 = *((UINT64 *)hp + 3);
    h5 = *((UINT64 *)hp + 4);
    h6 = *((UINT64 *)hp + 5);
    h7 = *((UINT64 *)hp + 6);
    h8 = *((UINT64 *)hp + 7);

    k0 = *(k+0); k1 = *(k+1); k2 = *(k+2); k3 = *(k+3); 
    k4 = *(k+4); k5 = *(k+5); k6 = *(k+6);

    do {
      d0 = LOAD_UINT32_LITTLE(d+0); d1 = LOAD_UINT32_LITTLE(d+1);
      d2 = LOAD_UINT32_LITTLE(d+2); d3 = LOAD_UINT32_LITTLE(d+3);

      k7 = *(k+7); k8 = *(k+8); k9 = *(k+9); k10 = *(k+10); 

      h1+= MUL64(d0,k0); h2+= MUL64(d0,k1); h3+= MUL64(d0,k2); h4+= MUL64(d0,k3); 
      h5+= MUL64(d0,k4); h6+= MUL64(d0,k5); h7+= MUL64(d0,k6); h8+= MUL64(d0,k7);

      h1+= MUL64(d1,k1); h2+= MUL64(d1,k2); h3+= MUL64(d1,k3); h4+= MUL64(d1,k4); 
      h5+= MUL64(d1,k5); h6+= MUL64(d1,k6); h7+= MUL64(d1,k7); h8+= MUL64(d1,k8);

      h1+= MUL64(d2,k2); h2+= MUL64(d2,k3); h3+= MUL64(d2,k4); h4+= MUL64(d2,k5); 
      h5+= MUL64(d2,k6); h6+= MUL64(d2,k7); h7+= MUL64(d2,k8); h8+= MUL64(d2,k9);

      h1+= MUL64(d3,k3); h2+= MUL64(d3,k4); h3+= MUL64(d3,k5); h4+= MUL64(d3,k6); 
      h5+= MUL64(d3,k7); h6+= MUL64(d3,k8); h7+= MUL64(d3,k9); h8+= MUL64(d3,k10);

      k0 = k4; k1 = k5; k2 = k6; k3 = k7; k4 = k8; k5 = k9; k6 = k10;
      d += 4;
      k += 4;
    } while (--c);
    ((UINT64 *)hp)[0] = h1;
    ((UINT64 *)hp)[1] = h2;
    ((UINT64 *)hp)[2] = h3;
    ((UINT64 *)hp)[3] = h4;
    ((UINT64 *)hp)[4] = h5;
    ((UINT64 *)hp)[5] = h6;
    ((UINT64 *)hp)[6] = h7;
    ((UINT64 *)hp)[7] = h8;
}


#endif


void modP(void *hp)
{
  UINT64 h, a, b, z, p;
  long y, c, d;

  p = (0x00010000UL) + 15;
  h = *((UINT64 *)hp);
  a = h >> 32;
  b = h & 0x0000ffffUL;
  y = b - 15 * a;

  if(y < 0){
    c = y >> 32;
    d = y & 0x0000ffffUL;
    z = d - 15 * c;
    if(z >= p){
      z = z - p;
    }
    *((UINT32 *)hp) = z;
  }else{
    *((UINT32 *)hp) = y;
  }
}

static double run_cpb_test(UINT8 mmh_key[], int nbytes, char *data_ptr,
                           int data_len, double hz)
{
    clock_t ticks;
    double secs;
    unsigned long iters_per_tag;
    unsigned long tag_iters, i, j, no_iters;
    unsigned long total_mbs;

    int len;
    char *input;
    UINT8 mmh_result[STREAMS*sizeof(UINT64)];

    total_mbs = 2500;
    tag_iters = (total_mbs * 1024 * 1024) / (nbytes) + 1;    
    iters_per_tag = nbytes / data_len;
    no_iters = iters_per_tag * tag_iters;

    //===================================================                
    len = data_len;
    input = data_ptr;
    while (len >= L1_KEY_LEN) {
      mmh_aux(mmh_key, input, mmh_result, L1_KEY_LEN);
      len -= L1_KEY_LEN;
      input += L1_KEY_LEN;
    }             
    //==================================================

    ticks = clock();
    for(i = 0; i < no_iters; i++){
	len = data_len;
	input = data_ptr;

	while (len >= L1_KEY_LEN) {
	  mmh_aux(mmh_key, input, mmh_result, L1_KEY_LEN);
	  len -= L1_KEY_LEN;
	  input += L1_KEY_LEN;
	}

	for(j = 0; j < STREAMS; j++){
	  modP((UINT64 *)mmh_result + j);
	}
    }
    ticks = clock() - ticks;
    secs = (double)ticks / CLOCKS_PER_SEC;

    return (secs * (hz/(tag_iters * nbytes)));
}

int main(void)
{
    aes_int_key prf_key;
    UINT8  mmh_key [L1_KEY_LEN + L1_KEY_SHIFT * (STREAMS - 1)];
    char *data_ptr;
    int data_len, i;
    double hz;
    double cpb;

    //int length_pts = 1024*1024; 
    int length_pts[] = {8*1024,32*1024,256*1024,1024*1024};
    
    //need to change the speed in Hz for each computer.
    hz = ((double)1e9);
    data_len = 8192;

    data_ptr = (char *)malloc(data_len + 16);

    for (i = 0; i < data_len; i++)
        data_ptr[i] = (i*i) % 128;

    aes_key_setup("abcdefghijklmnopqrstuvwxyz",prf_key);
    kdf(mmh_key, prf_key, 1, sizeof(mmh_key));
        
    //cpb = run_cpb_test(mmh_key, length_pts, data_ptr, data_len, hz);
    //printf("\nAuthenticating %8d byte messages: %5.2f cpb.\n", length_pts, cpb);

    printf("\nSpeed of %d-bit output MMH(k,m) \n",MMH_OUTPUT_LEN * 8);
    if (sizeof(length_pts) > 0) {
        for (i = 0; i < (int)(sizeof(length_pts)/sizeof(int)); i++) {
            cpb = run_cpb_test(mmh_key, length_pts[i], data_ptr, data_len, hz);
            printf("MMH hashing %8d byte messages: %5.2f cpb.\n",
                                   length_pts[i], cpb);
        }
    }
    printf("\n");
}




